import io
import re
from dataclasses import dataclass, field
from typing import Iterable, cast

import funcy
from pyld.jsonld import prepend_base
from requests import HTTPError, Session
from yarl import URL

from yaml_ld.document_loaders import content_types
from yaml_ld.document_loaders.base import DocumentLoader, DocumentLoaderOptions
from yaml_ld.errors import LoadingDocumentFailed, NotFound
from yaml_ld.models import URI, RemoteDocument

# Default `requests` timeout. Chosen arbitrarily.
DEFAULT_TIMEOUT = 30

LINK_PATTERN = re.compile(
    '<(?P<relative_link>[^>]+)>; rel="alternate"; '
    'type="(?P<content_type>[^"]+)"',
)


@dataclass
class LinkHeader:
    """Reference from HTTP Link header."""

    url: str
    rel: str
    content_type: str
    attributes: dict[str, str]


@funcy.post_processing(dict)
def _parse_link_attributes(
    raw_attributes: list[str],
) -> Iterable[tuple[str, str]]:
    for pair in raw_attributes:
        match_result = re.match(
            '(?P<attribute_name>[^=]+)="(?P<attribute_value>[^"]+)"',
            pair,
        )

        if match_result:
            yield cast(tuple[str, str], match_result.groups())
        else:
            raise ValueError(f'Cannot parse: {pair}')


def maybe_follow_one_of_link_headers(
    links: Iterable[LinkHeader],
    content_type: str | None,
    options: DocumentLoaderOptions,
):
    """Resolve the URL found in the given link headers (or not)."""
    from yaml_ld.document_loaders.default import (  # noqa: WPS433
        DEFAULT_DOCUMENT_LOADER,
    )

    for link in links:
        try:
            content_types.parser_by_content_type(
                content_type=link.content_type,
                uri=link.url,
            )
        except content_types.ParserNotFound:
            continue

        if not _is_content_type_more_preferable(   # noqa: WPS337
            left=link.content_type,
            right=content_type,
        ):
            continue

        return DEFAULT_DOCUMENT_LOADER(
            source=link.url,
            options=options,
        )

    return None


def parse_raw_link_header(   # noqa: WPS210
    page_url: str,
    link_header: str,
) -> Iterable[LinkHeader]:
    """Parse Link header into a structure."""
    links = link_header.split(', ')

    for link in links:
        bracketed_url, *raw_attributes = link.split('; ')
        relative_url = bracketed_url.removeprefix('<').removesuffix('>')

        absolute_url = prepend_base(page_url, relative_url)
        attributes: dict[str, str] = _parse_link_attributes(raw_attributes)

        if content_type := attributes.pop('type', None):
            yield LinkHeader(
                url=absolute_url,
                rel=attributes.pop('rel'),
                content_type=content_type,
                attributes=attributes,
            )


def _is_content_type_more_preferable(left: str, right: str | None) -> bool:
    """Determine if Left is better than Right."""
    if left == right:
        return False

    if left is not None and right is None:
        # Anything better than nothing
        return True

    return {
        ('application/rdf+xml', 'application/json'): False,
        ('application/rdf+xml', 'application/ld+json'): False,
        ('application/json', 'application/rdf+xml'): True,
        ('application/rdf+xml', 'text/html'): True,
        ('text/html', 'application/ld+json'): False,
        ('text/html', 'application/json'): False,
        ('application/ld+json', 'text/html'): True,
        ('application/json', 'text/html'): True,
        ('application/json', 'application/ld+json'): False,
    }[left, right]


@dataclass
class HTTPDocumentLoader(DocumentLoader):
    """Load documents from HTTP sources."""

    session: Session = field(default_factory=Session)

    def __call__(    # noqa: WPS210
        self,
        source: URI,
        options: DocumentLoaderOptions,
    ) -> RemoteDocument:
        """Load documents from HTTP sources."""
        string_source = str(source)
        url = URL(string_source)

        response = self.session.get(
            string_source,
            headers=options.get('headers'),
            timeout=DEFAULT_TIMEOUT,
        )

        try:
            response.raise_for_status()
        except HTTPError as http_error:
            match http_error.response.status_code:
                case 404:   # noqa: WPS432
                    raise NotFound(path=str(http_error.request.url))

            raise

        content_type = response.headers.get('Content-Type')

        if content_type is None:
            content_type = content_types.by_extension(url.suffix)

        if content_type is not None:
            content_type = re.sub(
                pattern='; charset=utf-8',
                repl='',
                string=content_type,
                flags=re.IGNORECASE,
            )

        if link := response.headers.get('Link'):
            follow_result = self.follow_link_header(
                source=source,
                content_type=content_type,
                link_header=link,
                options=options,
            )

            if follow_result:
                return follow_result

        if content_type is None:
            raise ValueError(
                f'What content type is extension `{url.suffix}`?',
            )

        parser = content_types.parser_by_content_type(
            content_type=content_type,
            uri=string_source,
        )
        if parser is None:
            raise LoadingDocumentFailed(path=source)

        yaml_document = parser(
            data_stream=io.BytesIO(response.content),
            source=string_source,
            options=options,
        )

        return {
            'document': yaml_document,
            'documentUrl': string_source,
            'contextUrl': None,
            'contentType': content_type,
        }

    def follow_link_header(
        self,
        source: URI,
        content_type: str | None,
        link_header: str,
        options: DocumentLoaderOptions,
    ) -> RemoteDocument | None:
        """Follow Link header."""
        links = parse_raw_link_header(
            page_url=str(source),
            link_header=link_header,
        )

        links = [
            link
            for link in links
            if link.rel == 'alternate'
        ]

        return maybe_follow_one_of_link_headers(
            links=links,
            content_type=content_type,
            options=options,
        )



from dataclasses import dataclass

from yaml_ld.document_loaders.base import DocumentLoader, DocumentLoaderOptions
from yaml_ld.models import URI, RemoteDocument


@dataclass
class MockLoader(DocumentLoader):
    """Mock loader returning the response specified as argument."""

    response: RemoteDocument

    def __call__(
        self,
        source: URI,
        options: DocumentLoaderOptions,
    ) -> RemoteDocument:
        """Return the prepared response."""
        return self.response



import functools
import json
import logging
import shutil
import sys
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from typing import Annotated, Optional

import funcy
import yaml
from documented import Documented
from rich.console import Console
from rich.errors import NotRenderableError
from rich.panel import Panel
from rich.syntax import Syntax
from typer import Argument, Option, Typer
from yarl import URL

import yaml_ld
from yaml_ld.compact import CompactOptions
from yaml_ld.document_loaders.default import CACHE_DIRECTORY
from yaml_ld.expand import ExpandOptions
from yaml_ld.flatten import FlattenOptions
from yaml_ld.from_rdf import FromRDFOptions
from yaml_ld.models import JsonLdRecord
from yaml_ld.to_rdf import ToRDFOptions

cli = Typer(
    no_args_is_help=True,
    help=(
        'Command line tool to operate on ＊-LD data, '
        'where ＊ stands for JSON or YAML.'
    ),
)
console = Console()


MaybeStr = Optional[str]


class LogLevel(StrEnum):
    """Logging level for the application."""

    DEBUG = 'debug'
    INFO = 'info'     # noqa: WPS110
    ERROR = 'error'


@cli.callback()
def _root_app_callback(  # noqa: WPS430
    log_level: Annotated[
        LogLevel, Option(help='Logging level.'),
    ] = LogLevel.ERROR,
):  # pragma: nocover
    cli.log_level = log_level   # type: ignore
    logging.basicConfig(
        level={
            LogLevel.ERROR: logging.ERROR,
            LogLevel.INFO: logging.INFO,
            LogLevel.DEBUG: logging.DEBUG,
        }[log_level],
    )


def decode_input(input_: str | None) -> Path | URL:
    """Interpret the input as location of the document."""
    match input_:
        case None:
            return Path('/dev/stdin')

        case str() as string:
            url = URL(string)

            if url.scheme:
                return url

            return Path(string)


class OutputFormat(StrEnum):
    """Output format."""

    JSON = 'json'
    YAML = 'yaml'


class RDFFormat(StrEnum):
    """Output format."""

    NQUADS = 'nquads'


def pretty_print(
    document: JsonLdRecord | list[JsonLdRecord],
    output_format: OutputFormat | RDFFormat,
) -> Syntax:
    """Serialize an LD document."""
    serializer = {
        OutputFormat.JSON: functools.partial(json.dumps, indent=2),
        OutputFormat.YAML: functools.partial(yaml.dump, Dumper=yaml.SafeDumper),
        RDFFormat.NQUADS: funcy.identity,
    }[output_format]

    serialized_document = serializer(document)

    return Syntax(
        serialized_document,
        lexer=output_format.value,
        background_color='default',
    )


@cli.command()
@funcy.post_processing(console.print)
def expand(
    input_: Annotated[
        MaybeStr,
        Argument(help='Path or URL. Omit to read from standard input.'),
    ] = None,
    output_format: Annotated[
        OutputFormat,
        Option(help='Format to output the data at.'),
    ] = OutputFormat.JSON,
    base: Annotated[MaybeStr, Option(help='Base URL.')] = None,
    extract_all_scripts: Annotated[
        bool,
        Option(
            help=(
                'Extract all documents in YAML stream, or all scripts '
                'embedded in HTML.'
            ),
        ),
    ] = False,
    expand_context: Annotated[
        MaybeStr,
        Option(help='Context to expand with.'),
    ] = None,
):
    """Expand a ＊-LD document."""
    response = yaml_ld.expand(
        document=decode_input(input_),
        options=ExpandOptions(
            base=base,
            extract_all_scripts=extract_all_scripts,
            expand_context=expand_context,
        ),
    )

    return pretty_print(
        document=response,
        output_format=output_format,
    )


@cli.command()
@funcy.post_processing(console.print)
def get(
    input_: Annotated[
        MaybeStr,
        Argument(help='Path or URL. Omit to read from standard input.'),
    ] = None,
    base: Annotated[MaybeStr, Option(help='Base URL.')] = None,
    output_format: Annotated[
        OutputFormat,
        Option(help='Format to output the data at.'),
    ] = OutputFormat.JSON,
):
    """Load and display a ＊-LD document."""
    response = yaml_ld.load_document(
        decode_input(input_),
        base=base,
    )['document']

    return pretty_print(
        document=response,
        output_format=output_format,
    )


@cli.command()
@funcy.post_processing(console.print)
def compact(   # noqa: WPS211
    input_: Annotated[
        MaybeStr,
        Argument(help='Path or URL. Omit to read from standard input.'),
    ] = None,
    ctx: Annotated[
        MaybeStr,
        Option(help='Context to compact with.'),
    ] = None,
    output_format: Annotated[
        OutputFormat,
        Option(help='Format to output the data at.'),
    ] = OutputFormat.JSON,
    base: Annotated[MaybeStr, Option(help='Base URL.')] = None,
    extract_all_scripts: Annotated[
        bool,
        Option(
            help=(
                'Extract all documents in YAML stream, or all scripts '
                'embedded in HTML.'
            ),
        ),
    ] = True,
    expand_context: Annotated[
        MaybeStr,
        Option(help='Context to expand with.'),
    ] = None,
):
    """Compact a ＊-LD document."""
    response = yaml_ld.compact(
        document=decode_input(input_),
        ctx=ctx or {},
        options=CompactOptions(
            base=base,
            extract_all_scripts=extract_all_scripts,
            expand_context=expand_context,
        ),
    )

    return pretty_print(
        document=response,
        output_format=output_format,
    )


@cli.command()
@funcy.post_processing(console.print)
def flatten(    # noqa: WPS211
    ctx: Annotated[
        MaybeStr,
        Option(help='Context to compact with.'),
    ] = None,
    input_: Annotated[
        MaybeStr,
        Argument(help='Path or URL. Omit to read from standard input.'),
    ] = None,
    output_format: Annotated[
        OutputFormat,
        Option(help='Format to output the data at.'),
    ] = OutputFormat.JSON,
    base: Annotated[MaybeStr, Option(help='Base URL.')] = None,
    extract_all_scripts: Annotated[
        bool,
        Option(
            help=(
                'Extract all documents in YAML stream, or all scripts '
                'embedded in HTML.'
            ),
        ),
    ] = True,
    expand_context: Annotated[
        MaybeStr,
        Option(help='Context to expand with.'),
    ] = None,
):
    """Flatten a ＊-LD document."""
    response = yaml_ld.flatten(
        document=decode_input(input_),
        ctx=ctx,
        options=FlattenOptions(
            base=base,
            extract_all_scripts=extract_all_scripts,
            expand_context=expand_context,
        ),
    )

    return pretty_print(
        document=response,
        output_format=output_format,
    )


@cli.command()
@funcy.post_processing(console.print)
def to_rdf(
    input_: Annotated[
        MaybeStr,
        Argument(help='Path or URL. Omit to read from standard input.'),
    ] = None,
    output_format: Annotated[
        RDFFormat,
        Option(help='Format to output the data at.'),
    ] = RDFFormat.NQUADS,
    base: Annotated[MaybeStr, Option(help='Base URL.')] = None,
    extract_all_scripts: Annotated[
        bool,
        Option(
            help=(
                'Extract all documents in YAML stream, or all scripts '
                'embedded in HTML.'
            ),
        ),
    ] = True,
):
    """Convert a ＊-LD document → RDF."""
    return yaml_ld.to_rdf(
        document=decode_input(input_),
        options=ToRDFOptions(
            base=base,
            format='application/n-quads',
            extract_all_scripts=extract_all_scripts,
        ),
    )


@cli.command()
@funcy.post_processing(console.print)
def from_rdf(
    input_: Annotated[
        MaybeStr,
        Argument(help='Path or URL. Omit to read from standard input.'),
    ] = None,
    output_format: Annotated[
        OutputFormat,
        Option(help='Format to output the data at.'),
    ] = OutputFormat.JSON,
):
    """Convert an RDF document → ＊-LD form."""
    response = yaml_ld.from_rdf(
        dataset=str(decode_input(input_)),
        options=FromRDFOptions(),
    )

    return pretty_print(
        document=response,
        output_format=output_format,
    )


cache = Typer(help='Cache management.', no_args_is_help=True)
cli.add_typer(cache, name='cache')


@cache.command()
def clear():
    """Clear cache."""
    shutil.rmtree(CACHE_DIRECTORY)
    console.print('Cache cleared.', style='green')


@dataclass
class FormattedError(Documented):   # type: ignore
    """**{self.exception_class}:** {self.message}"""  # noqa: D400

    exception: Exception

    @property
    def exception_class(self):
        """Class of the exception."""
        return self.exception.__class__.__name__

    @property
    def message(self):
        """Exception message."""
        return str(self.exception)


def print_unhandled_exception(err: Exception):  # pragma: no cover
    """Print unhandled exception as an error message."""
    try:
        console.print(Panel(err, style='red'))  # type: ignore
    except NotRenderableError:
        console.print(Panel(FormattedError(exception=err), style='red'))

    console.print(TracebackAdvice())


class TracebackAdvice(Documented):  # pragma: no cover
    """
    💡 To see Python traceback, use:

    `pyld --log-level info {self.args}`
    """  # noqa: D400

    @property
    def args(self):
        """Format current CLI args."""
        return ' '.join(funcy.rest(sys.argv))


def app() -> None:    # pragma: no cover
    """Construct and return Typer app."""
    try:
        return cli()

    except Exception as err:
        if cli.log_level == LogLevel.ERROR:    # type: ignore
            print_unhandled_exception(err)
            sys.exit(1)

        else:
            raise



import platformdirs
from requests_cache import CachedSession

from yaml_ld.document_loaders.choice_by_scheme import (
    ChoiceBySchemeDocumentLoader,
)
from yaml_ld.document_loaders.http import HTTPDocumentLoader
from yaml_ld.document_loaders.local_file import LocalFileDocumentLoader

CACHE_DIRECTORY = platformdirs.user_cache_dir(
    appname='python-yaml-ld',
    ensure_exists=True,
)

CACHED_SESSION = CachedSession(
    backend='filesystem',
    cache_name=CACHE_DIRECTORY,
)

http_loader = HTTPDocumentLoader(
    session=CACHED_SESSION,
)

DEFAULT_DOCUMENT_LOADER = ChoiceBySchemeDocumentLoader(
    file=LocalFileDocumentLoader(),
    http=http_loader,
    https=http_loader,
)



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.document_loaders.default import DEFAULT_DOCUMENT_LOADER
from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import DEFAULT_VALIDATE_CALL_CONFIG, JsonLdRecord
from yaml_ld.options import BaseOptions


class FromRDFOptions(BaseOptions):   # type: ignore
    """Options for converting RDF to ＊-LD."""

    format: str = 'application/n-quads'
    """The format if input is a string: 'application/n-quads' for N-Quads."""

    use_rdf_type: bool = False
    """True to use `rdf:type`, False to use `@type`."""

    use_native_types: bool = False
    """Convert XSD types into native types (boolean, integer, double)?"""


DEFAULT_FROM_RDF_OPTIONS = FromRDFOptions()


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def from_rdf(
    dataset: str,
    options: FromRDFOptions = DEFAULT_FROM_RDF_OPTIONS,
) -> JsonLdRecord:
    """Convert a RDF dataset to a [＊-LD](/blog/any-ld/) document."""
    dict_options = options.model_dump(by_alias=True, exclude_none=True)
    dict_options.setdefault('documentLoader', DEFAULT_DOCUMENT_LOADER)

    with except_json_ld_errors():
        return jsonld.from_rdf(dataset, dict_options)
